#ifndef H_LinkedListType  // Preprocessor directive used to avoid 
#define H_LinkedListType  // problems  resulting from multiple include.
                          

#include <iostream>
#include <cassert> // See assert function below
                  

using namespace std;

//Definition of the node

template <class Type>
struct nodeType
{
	Type info;
	nodeType<Type> *link;
};

template<class Type>
class linkedListType
{
public:
    const linkedListType<Type>& operator=
                         (const linkedListType<Type>&);  
      //Overload the assignment operator.
    void initializeList(); 
      //Initialize the list to an empty state.
      //Postcondition: first = NULL, last = NULL, count = 0;
    bool isEmptyList() const;
      //Function to determine whether the list is empty. 
      //Postcondition: Returns true if the list is empty,
      //               otherwise it returns false.
    void print() const;
      //Function to output the data contained in each node.
      //Postcondition: none
    int length() const;
      //Function to return the number of nodes in the list.
      //Postcondition: The value of count is returned.
    void destroyList();
      //Function to delete all the nodes from the list.
      //Postcondition: first = NULL, last = NULL, count = 0;
    Type front() const; 
      //Function to return the first element of the list.
      //Precondition: The list must exist and must not be 
      //              empty.
      //Postcondition: If the list is empty, the program
      //               terminates; otherwise, the first 
      //               element of the list is returned.
    Type back()const; 
      //Function to return the last element of the list.
      //Precondition: The list must exist and must not be 
      //              empty.
      //Postcondition: If the list is empty, the program
      //               terminates; otherwise, the last  
      //               element of the list is returned.

    bool search(const Type& searchItem) const;
      //Function to determine whether searchItem is in the list.
      //Postcondition: Returns true if searchItem is in the 
      //               list, otherwise the value false is 
      //               returned.

    void insertFirst(const Type& newItem);
      //Function to insert newItem at the beginning of the list.
      //Postcondition: first points to the new list, newItem is
      //               inserted at the beginning of the list,
      //               last points to the last node in the list, 
      //               and count is incremented by 1.

    void insertLast(const Type& newItem);
      //Function to insert newItem at the end of the list.
      //Postcondition: first points to the new list, newItem 
      //               is inserted at the end of the list,
      //               last points to the last node in the list,
      //               and count is incremented by 1.

    void deleteNode(const Type& deleteItem);
      //Function to delete deleteItem from the list.
      //Postcondition: If found, the node containing 
      //               deleteItem is deleted from the list.
      //               first points to the first node, last
      //               points to the last node of the updated 
      //               list, and count is decremented by 1.


    linkedListType(); 
      //default constructor
      //Initializes the list to an empty state.
      //Postcondition: first = NULL, last = NULL, count = 0; 

    linkedListType(const linkedListType<Type>& otherList); 
         //copy constructor

    ~linkedListType();   
      //destructor
      //Deletes all the nodes from the list.
      //Postcondition: The list object is destroyed. 

	//ADDED BY ASSIGNMENT
	Type extractMin();
	void divideMid(linkedListType<Type>& subList);
	void mergeLists(linkedListType<Type> & List1, linkedListType<Type> & List2);
	int getCount();

	//ADDED BY FINAL EXAM
	void removeDuplicates();

protected:
    int count;   //variable to store the number of 
                 //elements in the list
    nodeType<Type> *first; //pointer to the first node of the list
    nodeType<Type> *last;  //pointer to the last node of the list

private: 
    void copyList(const linkedListType<Type>& otherList); 
      //Function to make a copy of otherList.
      //Postcondition: A copy of otherList is created and
      //               assigned to this list.
};


template<class Type>
bool linkedListType<Type>::isEmptyList() const
{
	return(first == NULL);
}

template<class Type>
linkedListType<Type>::linkedListType() //default constructor
{
	first = NULL;
	last = NULL;
 	count = 0;
}

template<class Type>
void linkedListType<Type>::destroyList()
{
    nodeType<Type> *temp;   //pointer to deallocate the memory
                            //occupied by the node
    while (first != NULL)   //while there are nodes in the list
    {
        temp = first;        //set temp to the current node
        first = first->link; //advance first to the next node
        delete temp;        //deallocate the memory occupied by temp
    }
    last = NULL; //initialize last to NULL; first has already
                 //been set to NULL by the while loop
    count = 0;
}
	
template<class Type>
void linkedListType<Type>::initializeList()
{
	destroyList(); //if the list has any nodes, delete them
}

template<class Type>
void linkedListType<Type>::print() const
{
	nodeType<Type> *current; //pointer to traverse the list

	current = first;  //set current so that it points to 
       			      //the first node
	while (current != NULL) //while more data to print
	{
	   cout << current->info << " ";
	   current = current->link;
	}
}//end print


template<class Type>
int linkedListType<Type>::length() const
{
    return count;
}  //end length

template<class Type>
Type linkedListType<Type>::front() const
{   
    assert(last != NULL);  // terminate program if last!=NULL  
	// is essentially the same as: if(last==NULL) exit(0);
    return first->info; //return the info of the first node	
}//end front

template<class Type>
Type linkedListType<Type>::back() const
{   
    assert(last != NULL); // terminate program if last!=NULL (see [Malik, page 187]) 
	// is essentially the same as: if(last==NULL) exit(0);
    return last->info; //return the info of the first node	
}//end back


template<class Type>
bool linkedListType<Type>::search(const Type& searchItem) const
{
    nodeType<Type> *current; //pointer to traverse the list
    bool found = false;
    
    current = first; //set current to point to the first 
                     //node in the list

    while (current != NULL && !found)    //search the list
        if (current->info == searchItem) //searchItem is found
            found = true;
        else
            current = current->link; //make current point to
                                     //the next node
    return found; 
}//end search

template<class Type>
void linkedListType<Type>::insertFirst(const Type& newItem)
{
   nodeType<Type> *newNode; //pointer to create the new node

   newNode = new nodeType<Type>; //create the new node

   assert(newNode != NULL);      //if unable to allocate memory, 
                                 //terminate the program (see [Malik, page 187]) 

   newNode->info = newItem; 	   //store the new item in the node
   newNode->link = first;        //insert newNode before first
   first = newNode;              //make first point to the 
                                 //actual first node
   count++; 			   //increment count

   if (last == NULL)   //if the list was empty, newNode is also 
                      //the last node in the list
      last = newNode;
}//end insertFirst

template<class Type>
void linkedListType<Type>::insertLast(const Type& newItem)
{
   nodeType<Type> *newNode; //pointer to create the new node

   newNode = new nodeType<Type>; //create the new node

   assert(newNode != NULL);	//if unable to allocate memory, 
 					//terminate the program (see [Malik, page 187]) 

   newNode->info = newItem;      //store the new item in the node
   newNode->link = NULL;         //set the link field of newNode
         				   //to NULL

   if (first == NULL)  //if the list is empty, newNode is 
     			    //both the first and last node
   {
	  first = newNode;
	  last = newNode;
	  count++;		//increment count
   }
   else     //the list is not empty, insert newNode after last
   {
	last->link = newNode; //insert newNode after last
	last = newNode;   //make last point to the actual last node
      count++;		//increment count
   }
}//end insertLast


template<class Type>
void linkedListType<Type>::deleteNode(const Type& deleteItem)
{
    nodeType<Type> *current; //pointer to traverse the list
    nodeType<Type> *trailCurrent; //pointer just before current
    bool found;

    if (first == NULL)    //Case 1; the list is empty. 
        cout << "Cannot delete from an empty list."
             << endl;
    else
    {
        if (first->info == deleteItem) //Case 2 
        {
            current = first;
            first = first->link;
            count--;
            if (first == NULL)    //the list has only one node
                last = NULL;
            delete current;
        }
        else //search the list for the node with the given info
        {
            found = false;
            trailCurrent = first;  //set trailCurrent to point
                                   //to the first node
            current = first->link; //set current to point to 
                                   //the second node

            while (current != NULL && !found)
            {
                if (current->info != deleteItem) 
                {
                    trailCurrent = current;
                    current = current-> link;
                }
                else
                    found = true;
            }//end while

            if (found) //Case 3; if found, delete the node
            {
                trailCurrent->link = current->link;
                count--;

                if (last == current)   //node to be deleted 
                                       //was the last node
                    last = trailCurrent; //update the value 
                                         //of last
                delete current;  //delete the node from the list
            }
            else
                cout << "The item to be deleted is not in "
                     << "the list." << endl;
        }//end else
    }//end else
}//end deleteNode

template<class Type>
void linkedListType<Type>::copyList
                   (const linkedListType<Type>& otherList) 
{
    nodeType<Type> *newNode; //pointer to create a node
    nodeType<Type> *current; //pointer to traverse the list

    if (first != NULL) //if the list is nonempty, make it empty
       destroyList();

    if (otherList.first == NULL) //otherList is empty
    {
        first = NULL;
        last = NULL;
        count = 0;
    }
    else
    {
        current = otherList.first; //current points to the 
                                   //list to be copied
        count = otherList.count;

            //copy the first node
        first = new nodeType<Type>;  //create the node

        assert(first != NULL); 

        first->info = current->info; //copy the info
        first->link = NULL;        //set the link field of 
                                   //the node to NULL
        last = first;              //make last point to the
                                   //first node
        current = current->link;     //make current point to
                                     //the next node

           //copy the remaining list
        while (current != NULL)
        {
            newNode = new nodeType<Type>;  //create a node

            assert(newNode != NULL);

            newNode->info = current->info; //copy the info
            newNode->link = NULL;       //set the link of 
                                        //newNode to NULL
            last->link = newNode;  //attach newNode after last
            last = newNode;        //make last point to
                                   //the actual last node
            current = current->link;   //make current point 
                                       //to the next node
        }//end while
    }//end else
}//end copyList

template<class Type>
linkedListType<Type>::~linkedListType() //destructor
{
   destroyList();
}//end destructor

template<class Type>
linkedListType<Type>::linkedListType
            	    (const linkedListType<Type>& otherList) 
{
   	first = NULL;
    copyList(otherList);
}//end copy constructor

         //overload the assignment operator
template<class Type>
const linkedListType<Type>& linkedListType<Type>::operator=
                      (const linkedListType<Type>& otherList)
{ 
    if (this != &otherList) //avoid self-copy
    {
        copyList(otherList);
    }//end else

     return *this; 
}

template<class Type>
Type linkedListType<Type>::extractMin() {
	if (first !=NULL){
		nodeType<Type> *current = first;
		Type min = current->info;
		for (int i=1;i<=count;i++) {
			if (current->info < min) {
				min = current->info;	
			}
			current = current->link;
		}
		deleteNode(min);
		return (min);
	}
	else exit(0);
}

template<class Type>
void linkedListType<Type>::divideMid(linkedListType<Type>& subList) {
	if (count > 1) {
		int count1 , tempCount=count;
		if (count%2==0) count1 = count/2;
		else count1=count/2+1;
		
		nodeType<Type> *current = first;
		for (int i=1;i<=count1;i++) {
			last = current;
			current = current->link;
		}

		subList.destroyList();
		
		subList.count=count-count1;
		int subCount=subList.count;
		

		for (i=1;i<=subCount;i++) {
			subList.insertLast(current->info);
			nodeType<Type> *temp = current;
			current= current->link;
			deleteNode(temp->info);
		}
		subList.count=tempCount-count1;
		count=count1;
	}
}

template<class Type>
void linkedListType<Type>::mergeLists(linkedListType<Type> & List1, linkedListType<Type> & List2) 
{
	destroyList();
	int i =1, j = 1 , m=List1.count, n=List2.count;
	nodeType<Type> *currentL = List1.first;
	nodeType<Type> *currentR = List2.first;
	
	count=m+n;
	int maxSize=count;
	
	for(int k = 1;k<=maxSize;k++)
	{
		if (i<= m && j <= n) {
			Type iL = currentL->info, iR = currentR->info;
			
			if ( iL < iR ) { 
				insertLast(currentL->info); 
				if (currentL != List1.last) currentL = currentL->link;
				i++;
			}
			else { 
				insertLast(currentR->info); 
				if (currentR != List2.last) currentR = currentR->link; 
				j++;
			}
		}
		else if (i==m+1) {
			insertLast(currentR->info); 
			if (currentR != List2.last) currentR = currentR->link;
			j++;
		} 
		else if (j==n+1) {
			insertLast(currentL->info);
			if (currentL != List1.last) currentL = currentL->link;
			i++;
		}
	}
	List1.destroyList();
	List2.destroyList();
	count=m+n;
}

template<class Type>
int linkedListType<Type>::getCount() {
	return count;
}

template<class Type>
void linkedMergeSort(linkedListType<Type> & L) {
	int tempCountL = L.getCount();
	if(tempCountL>1)  {
		linkedListType<Type> L2, L3;
		L.divideMid(L2);
		linkedMergeSort(L);
		linkedMergeSort(L2);
		L3.mergeLists(L,L2);
		L=L3;
	} 
}

template<class Type>
void linkedListType<Type>::removeDuplicates(){
	int max = count;
	nodeType<Type> *current = first;
	Type temp = current->info;
	current = current->link;
	for (int i=1;i<max;i++) {
		Type curInfo = current->info;
		if (curInfo == temp) {
			deleteNode(curInfo);
		}
		temp = current ->info;
		current = current->link;
	}
}
#endif





















